#include <msp430.h> 

/*
 * main.c

	Colin McGlothlin
	August 20, 2014
	
	Bubble Clock Watch
 
 */

#define DATA BIT0
#define LATCH BIT1
#define SHIFT BIT2
#define DIG1 BIT3
#define DIG2 BIT4
#define DIG3 BIT5
#define DIG4 BIT6

/* Variables which are updated within an interrupt routine have to be "volatile" so that
the compiler knows their values can change at any time. */
volatile int halfSeconds;
volatile int seconds;
volatile int minutes;
volatile int hours;
volatile int timeout;
volatile int showTime;
volatile int buttonPressed;
volatile int buttonReleased;
volatile int digit1;
volatile int digit2;
volatile int digit3;
volatile int digit4;
int buttonTime;
int setHours;
int setMinutes;
int timeMode;
int buttonDown;
volatile int buttonCounter;

/* 7-segment codes for each digit, 0-9 and null */
unsigned char digit[] = {
    0x7E,	  0x30,	  0x6D,	  0x79,	  0x33,	  0x5B,	  0x5F,	  0x70,	  0xFF,	  0x7B,   0x00};
//  0       1       2       3       4       5       6       7       8       9       null
//	1111110	0110000	1101101	1111001	0110011	1011011	1011111	1110000	1111111	1111011	0000000
//	abcdefg	abcdefg	abcdefg	abcdefg	abcdefg	abcdefg	abcdefg	abcdefg	abcdefg	abcdefg	abcdefg
//  Normal (common cathode) codes are above
//  Inverse (common anode) combinations are below
//	0x01, 	0x4F,	  0x12,	  0x06,	  0x4C,	  0x24,	  0x20,	  0x0F,	  0x00,	  0x04,   0xFF};
//  0       1       2       3       4       5       6       7       8       9       null
//	0000001	1001111	0010010 0000110 1001100 0100100 0100000 0001111 0000000 0000100	1111111
/*
		 ---a---
		|       |
		f       b
		|       |
		 ---g---
		|       |
		e       c
		|       |
		 ---d---   .
*/


/* function sendDigit
	This function sends the digit segment data to the shift register for output on the display.
	
	There are two inputs:
	
	d, an integer from 0 to 10, representing the character to be displayed. A value of 10 means
	the null character (no segments should be lit at all).
	
	decimal, a boolean value representing the decimal point.  A value of 1 means to light the
	decimal point for that digit, and 0 means don't light it.

*/
void sendDigit(unsigned int d, unsigned int decimal) {

	P1OUT |= DATA & decimal;			// If decimal is 1, set the data pin HIGH; if 0, leave it alone.
	P1OUT |= SHIFT;						// Set the shift pin HIGH
	P1OUT &= ~SHIFT;					// Set the shift pin LOW
	P1OUT &= ~DATA;						// Set the data pin LOW (whether it was HIGH or LOW)
	P1OUT |= DATA && (BIT0 & digit[d]);	// If the "a" segment for the desired digit should be on, set the data pin HIGH; if not, leave it alone.
	P1OUT |= SHIFT;						// Set the shift pin HIGH
	P1OUT &= ~SHIFT;					// Set the shift pin LOW
	P1OUT &= ~DATA;						// Set the data pin LOW (whether it was HIGH or LOW)
	P1OUT |= DATA && (BIT1 & digit[d]);	// If the "b" segment...
	P1OUT |= SHIFT;						
	P1OUT &= ~SHIFT;					
	P1OUT &= ~DATA;						
	P1OUT |= DATA && (BIT2 & digit[d]);	// If the "c" segment...
	P1OUT |= SHIFT;						
	P1OUT &= ~SHIFT;					
	P1OUT &= ~DATA;						
	P1OUT |= DATA && (BIT3 & digit[d]);	// If the "d" segment...
	P1OUT |= SHIFT;						
	P1OUT &= ~SHIFT;					
	P1OUT &= ~DATA;						
	P1OUT |= DATA && (BIT4 & digit[d]);	// If the "e" segment...
	P1OUT |= SHIFT;						
	P1OUT &= ~SHIFT;					
	P1OUT &= ~DATA;						
	P1OUT |= DATA && (BIT5 & digit[d]);	// If the "f" segment...
	P1OUT |= SHIFT;						
	P1OUT &= ~SHIFT;					
	P1OUT &= ~DATA;						
	P1OUT |= DATA && (BIT6 & digit[d]);	// If the "g" segment...
	P1OUT |= SHIFT;						
	P1OUT &= ~SHIFT;					
	P1OUT &= ~DATA;						
}

/* function latchDigit

	This function sends the latch (update) signal to the shift register by pulsing the latch pin.  This updates the
	display to whatever data was just sent to the shift register.
*/
void latchDigit() {
	P1OUT |= LATCH;		// Set the latch pin HIGH
	P1OUT &= ~LATCH;	// Set the latch pin LOW
}

/*	function incrementTime
	
	This function increments the time, and is called every second.	
	This function operates in 12-hour time.	
	I think this function can be modified to operate in 24-hour time quite easily.  Just change
	"(hours > 12)" to "(hours > 23)" and "hours = 1" to "hours = 0"
	
*/
void incrementTime() {
	seconds++;
	if (seconds > 59) {			// If we have reached 60 seconds,
		seconds = 0;			// Reset seconds to zero
		minutes++;				// Add one minute to the minute count
		if (minutes > 59) {		// If we have reached 60 minutes,
			minutes = 0;		// Reset minutes to zero
			hours++;			// Add one hour to the hour count
			if (hours > 12) {	// If we have reached the end of a 12-hour period,
				hours = 1;		// Reset hours to one
			}
		}
	}
}

/*	function formatTime

	This function takes the current minutes and hours and translates them into four digits which can
	then be sent to the display.
	
	The digits are formatted this way:
	
	time    HH.MM
	digits  12 34
	
	The modulo operator "%" and division operator "/"is a handy way to break an integer, say 25,
	into 2 and 5.  25%10 returns the remainder of 25/10, which is 5.  25/10 divides 25 by 10, which is 2,
	not 2.5 because it's an integer.

*/
void formatTime() {
	digit4 = minutes % 10;				// Digit4 is set to the minutes' "ones" digit.  
	digit3 = minutes / 10;				// Digit3 is set to the minutes' "tens" digit.
	digit2 = hours % 10;				// Digit2 is set to the hours' "ones" digit.
	
	
/*	What the following lines do is blank out the "tens" digit if the number is zero.  Using 12-hour time,
	we are used to seeing one o'clock as 1:00, and not 01:00.  Otherwise, things would be confusing once
	the afternoon comes and all clocks would look like a 24-hour clock that is 12 hours off. */
	int temp = hours / 10;				// Divide hours by 10 to get the hours' "tens"
	if (temp == 0)						// If the "tens" digit is zero,
		digit1 = 10;					// Hide the digit (display character 10, which is totally blank)
	else								// If the "tens" digit is not zero,
		digit1 = temp;					// Display the first digit.
		
	/* The last 4 lines above can be written as one:
	
	digit1 = temp + (!temp  * 10);

	In C, an integer can be treated as boolean; 0 means false, and any value not equal to 0 is true.  So,
	assuming temp = 0, !temp returns true, since !false is true.  Likewise, if temp = 2, !temp would be false.
	
	*/
}

/*	function incrementSeconds
	This function allows the seconds to keep ticking while the clock is in set mode.  Seconds keep ticking and
	rolling over to 0, but the minutes do not advance after 60 seconds.

	The watch sets the seconds to zero each time minutes are set.
*/
void incrementSeconds() {
	seconds++;
	if (seconds > 59)
		seconds = 0;
}

/*	function setupPins

	This function sets up the pins on the MSP430.  This code is run when the MSP430 is first powered on, and also
	when the pushbutton is pressed.
*/
void setupPins() {
	/* Set up the pin direction register.  |= means we are selectively turning on bits in the register, defining
	the output pins. The values in all caps are defined at the very top of this file. */
	P1DIR |= DATA + LATCH + SHIFT + DIG1 + DIG2 + DIG3 + DIG4;
	/* Set up the pin direction register.  &= ~ means we are selectively turning off bits in the register, defining
	any input pins.  The pushbutton is connected to pin 9, which corresponds to BIT7 on the pin register.*/
	P1DIR &= ~BIT7;
	/* Set the pins to LOW that are connected to the shift register. */
	P1OUT &= ~(DATA + LATCH + SHIFT);
	/* Set the pins to HIGH that are connected to the cathode of each digit. */
	P1OUT |= DIG1 + DIG2 + DIG3 + DIG4;
}

/*	function sleepPins
	
	This function sets all of the used pins to input (high-impedance) and LOW, to minimize current while
	the MSP430 is asleep.
*/
void sleepPins() {
	P1DIR &= ~(DATA + LATCH + SHIFT + DIG1 + DIG2 + DIG3 + DIG4 + BIT7);
	P1OUT &= ~(DATA + LATCH + SHIFT + DIG1 + DIG2 + DIG3 + DIG4);
}


int main(void) {
	/* Stop the watchdog timer. */
	WDTCTL = WDTPW | WDTHOLD;
	
	/* Enable clock interrupts */
	CCTL0 = CCIE;
	/* Set the clock to count up 16384 cycles (1/2 second) before sending an interrupt.  The -1 is there because the
	clock starts its count at zero, and it's easier to do that rather than always in your head have to 	subtract 1
	from your desired count. */
	CCR0 = 16384-1;
	/* Tell the timer controller to count using the external 32.768 kHz crystal.  TASSEL_1 corresponds to the external
	crystal.  MC_1 tells the timer controller to count in "upmode", which counts up to the designated value (16384-1),
	resets to zero, and starts counting up again. */
	TACTL = TASSEL_1 + MC_1;				// ACLK (Auxilliary Clock), upmode

	/* Enable pin interrupt on pin 9 (BIT7) */
	P1IE |= BIT7;
	/* Set pin 9 to trigger an interrupt on the falling edge.  (P1IES = Port 1 Interrupt Edge Select)*/
	P1IES |= BIT7;
	/* Clear the interrupt flag for pin 9. */
	P1IFG &= ~BIT7;

	/* Set up the pin directions (input/output) and their initial output values. */
	setupPins();

	/* Initialize the clock time at 12:00:00 Noon/Midnight.  The clock does not care which. */
	halfSeconds = 0;
	seconds = 0;
	minutes = 0;
	hours = 12;

//	timeout = 3;
//	buttonReleased = 0;

	/* Initialize the state variables */
	/* showTime is a counter in half-seconds.  When the button is pressed, the display lights up,the counter gets set
	to 10 (5 seconds), and counts down to zero, and then the display turns off. */
	showTime = 0;
	/* buttonPressed is a boolean state variable that indicates whether or not the button is currently pressed. */
	buttonPressed = 0;
	/* buttonTime is a counter in half-seconds.  When the button is held down (not just pressed), the counter starts
	counting up. When a certain number is reached, set mode is activated.*/
	buttonTime = 0;
	/* setHours is a boolean state variable that is set to true when the hours are being set. */
	setHours = 0;
	/* setMinutes is a boolean state variable that is set to true when the minutes are being set. */
	setMinutes = 0;
	/* timeMode is a boolean state variable that is set to true when the time is being displayed. */
	timeMode = 0;
	/* buttonDown records a button press and does not reset when the button is released.  This allows logic to happen
	after the button is released and buttonPressed has returned to false. */
	buttonDown = 0;
	
	/* Initialize each display digit. 10 means to leave the digit blank. */
	digit1 = 10;	// hours' "tens"
	digit2 = 10;	// hours' "ones"
	digit3 = 10;	// minutes' "tens"
	digit4 = 10;	// minutes' "ones"

	/* Enable Low Power Mode 3 (LPM3) with global interrupts enabled (GIE).  LPM3 is the lowest power mode where the
	timers and interrupts are still active.  Without GIE, the clock will never "tick" when the MSP430 is asleep. 
	__bis is for setting the bits and entering low power mode, while __bic is for clearing the bits and waking up.
	s for set, c for clear.*/
	__bis_SR_register(LPM3_bits + GIE);		

	/* Infinite loop while the MSP430 is awake. */
	while(1) {

		if (timeMode) {													// If the time is currently being displayed,
			/* Since the 4-digit 7-segment display is multiplexed, only one digit can be displayed at a time.  Therefore,
			the MSP430 has to flash all 4 digits constantly.
			
			1) Send a new digit while the old one is still displaying.
			2) Turn off the old digit.
			3) Switch to the new digit.
			4) Display the new digit.
			
			*/

//			if(showTime > 0 || (buttonPressed && buttonTime > 9)) {		// If there is time left on the showTime counter OR (the button is pressed AND the button has been held for more than 4.5 seconds),
			if (buttonPressed) {
				sendDigit(digit4, 0);									// Send the minutes' "ones" digit.
				P1OUT |= DIG1;											// Turn off the old digit.
				latchDigit();											// Tell the shift register to update to the new digit.
				P1OUT &= ~DIG4;											// Show the new digit.
				
				sendDigit(digit3, 0);									// Send the minutes' "tens" digit.
				P1OUT |= DIG4;											// Turn off the old digit.
				latchDigit();											// Update...
				P1OUT &= ~DIG3;											// Show the new digit.
				
				sendDigit(digit2, !halfSeconds);						// Send the hours' "ones" digit.
				P1OUT |= DIG3;											// Turn off the old digit.
				latchDigit();											// Update...
				P1OUT &= ~DIG2;											// Show the new digit.
				
				sendDigit(digit1, 0);									// Send the hours' "tens" digit.
				P1OUT |= DIG2;											// Turn off the old digit.
				latchDigit();											// Update...
				P1OUT &= ~DIG1;											// Show the new digit.
			}
			else { // time expired, button released				
				P1OUT |= DIG1;									// Turn off the display,
				timeMode = 0;									// set timeMode to false,
				buttonPressed = 0;								// set buttonPressed to false,
				buttonTime = 0;									// set buttonTime to zero,
				sleepPins();									// put the pins to sleep, and
				__bis_SR_register(LPM3_bits + GIE);				// activate LPM3 sleep mode with interrupts enabled.
			}

			if (buttonTime > 20) {						// If the button has been held down for more than 10 seconds,
				buttonPressed = 0;						// Set buttonPressed to false.
				buttonTime = 0;							// Reset buttonTime to zero.
				// enter set mode						
				timeMode = 0;							// Turn off time display mode.
				setHours = 1;							// Turn on set hours mode.
				buttonDown = 0;							// Set buttonDown to false.
			} else {								// If the button hasn't been held down long enough
				if (buttonPressed == 0 || (P1IN & BIT7)){				// and if the button was just released,
					buttonPressed = 0;				// make sure to set buttonPressed to false, and
					buttonTime = 0;					// reset buttonTime to zero.
				}
			}
		}
		if (setHours) {										//	If hours set mode is active,
			if (buttonTime > 5) {							//		If the button has been held for longer than 2.5 seconds,
				buttonTime = 0;								//			reset buttonTime,
				buttonPressed = 0;							//			reset buttonPressed,
				buttonDown = 0;								//			reset buttonDown,
				setHours = 0;								//			turn off hours set mode, and
				setMinutes = 1;								//			activate minutes set mode.
			} else {										//		If the button hasn't been held down long enough,
				if (buttonPressed == 0 || (P1IN & BIT7)) {	//			and if the button was just released,
					buttonPressed = 0;						//				reset buttonPressed
					buttonTime = 0;							//				reset buttonTime
				}
			}
			if (buttonPressed)
				buttonDown = 1;
			else {
				if (buttonDown && buttonTime < 2) {
					hours++;
					if (hours > 12)
						hours = 1;
					buttonDown = 0;
					formatTime();
				}
			}

			sendDigit(digit4, 0);
			P1OUT |= DIG1;
			latchDigit();
			P1OUT &= ~DIG4;
			
			sendDigit(digit3, 0);
			P1OUT |= DIG4;
			latchDigit();
			P1OUT &= ~DIG3;
			
			sendDigit(digit2, 1);
			P1OUT |= DIG3;
			latchDigit();
			P1OUT &= ~DIG2;
			
			sendDigit(digit1, 1);				// In hour set mode, the first digit's decimal point is turned on, like this: 1.2.0 0
			P1OUT |= DIG2;
			latchDigit();
			P1OUT &= ~DIG1;
		}
		if (setMinutes) {



			sendDigit(digit4, 1);				// In minute set mode, the fourth digit's decimal point is turned on, like this: 1 2.0.0
			P1OUT |= DIG1;
			latchDigit();
			P1OUT &= ~DIG4;
			sendDigit(digit3, 0);
			P1OUT |= DIG4;
			latchDigit();
			P1OUT &= ~DIG3;
			sendDigit(digit2, 1);
			P1OUT |= DIG3;
			latchDigit();
			P1OUT &= ~DIG2;
			sendDigit(digit1, 0);
			P1OUT |= DIG2;
			latchDigit();
			P1OUT &= ~DIG1;

			if (buttonTime > 5) {
				buttonTime = 0;
				buttonPressed = 0;
				buttonDown = 0;
				setMinutes = 0;
				timeMode = 0;
				sleepPins();
				__bis_SR_register(LPM3_bits + GIE);
			} else {
				if (buttonPressed == 0)
					buttonTime = 0;
			}
			if (buttonPressed)
				buttonDown = 1;
			else {
				if (buttonDown && buttonTime < 2) {

					TACTL = 0;
					__delay_cycles(10);
					TACTL = TASSEL_1 + MC_1;
					minutes++;
					seconds = 0;
					halfSeconds = 0;
					if (minutes > 59)
						minutes = 0;
					buttonDown = 0;
					formatTime();
				}
			}
		}
	}

	return 0;
}

/* Timer0 A0 interrupt service routine
	This function runs once every half-second, triggered by
	the 32.768 kHz crystal connected to pins 12 and 13.
*/
#pragma vector=TIMER0_A0_VECTOR
__interrupt void Timer_A (void)
{
	if (buttonPressed) {					// If the button is currently being pressed,
		buttonTime++;						// then increment buttonTime by 1.
	}
	if (showTime > 0) {						// If there is time left for showing the current time,
		showTime--;							// then decrement showTime by 1.
	}
	halfSeconds++;							// Increment halfSeconds by 1.
	if (halfSeconds > 1) {					// If halfSeconds is greater than 1,
		halfSeconds = 0;					// then reset halfSeconds to 0,
		if (!(setHours || setMinutes)) {	// and if both setHours and setMinutes are false,
			incrementTime();				// then run the function incrementTime (add 1 second to the current time)
			if (timeMode)					// and if the time is currently being displayed,
				formatTime();				// then update the digits on the display.
		}
		else
			incrementSeconds();
	}
}

/*	Button interrupt service routine 
	This function runs when the button is both pressed and released.
	
	If the MSP430 is asleep, pressing the button wakes it temporarily, enough to run this function.

*/
#pragma vector=PORT1_VECTOR
__interrupt void Port_1(void)
{
	if (P1IES & BIT7) {									// If BIT7 in the Interrupt Edge Select register (P1IES) is true (therefore detecting a button press),
		buttonPressed = 1;								// then set buttonPressed to true.
		if (!(timeMode || setMinutes || setHours)) {	// And, if timeMode, setMinutes, and setHours are all false (nothing is currently being displayed),
			timeMode = 1;								// then set timeMode to true,
			showTime = 10;								// and set the showTime counter to 10 (10 half-seconds, or 5 seconds).
		}

		formatTime();									// Update the digits on the display.
		setupPins();									// Prepare the MSP430 to talk to the shift register, and to run the display.
		__bic_SR_register_on_exit(LPM3_bits);			// Tell the MSP430 to leave LPM3 and awaken fully when this function is done.
	} else {											// If BIT7 in the Interrupt Edge Select register is false (therefore detecting a button release),
		buttonPressed = 0;								// then set buttonPressed to false.
	}

	/* Toggle the Interrupt Edge Select for BIT7.  This allows the same function to detect both a button press and then a button release. */
	P1IES ^= BIT7;
	/* Clear the Interrupt Flag for BIT7. */
	P1IFG &= ~BIT7;
}
